import torch as th
from typing import Optional

import hashlib

from mamo.api_support import (
    ExternallyCachedValue,
    ModuleExtension,
    ObjectSaver,
    ExternallyCachedFilePath,
    CachedValue,
    DBPickledValue,
    MODULE_EXTENSIONS,
)


# TODO: does this all work for cuda tensors????


class TorchExternallyCachedValue(ExternallyCachedValue):
    def load(self):
        return th.load(self.path)

    @staticmethod
    def save(external_path, value):
        th.save(value, external_path)

        return TorchExternallyCachedValue(external_path)


class TorchObjectSaver(ObjectSaver):
    def __init__(self, value: th.Tensor):
        super().__init__(value)
        self.value = value

    def get_estimated_size(self) -> Optional[int]:
        return self.value.numel() * self.value.element_size()

    def compute_digest_(self):
        return hashlib.md5(self.value.numpy()).digest()

    def cache_value(self, external_path_builder: Optional[ExternallyCachedFilePath]) -> Optional[CachedValue]:
        if external_path_builder is None:
            return DBPickledValue.cache_value(self.value)

        shape_info = "_".join(map(str, self.value.shape))
        external_path = external_path_builder.build(shape_info, "pth")

        return TorchExternallyCachedValue.save(external_path, self.value)


class TorchModuleExtension(ModuleExtension):
    def supports(self, value) -> bool:
        return isinstance(value, th.Tensor)

    def get_object_saver(self, value) -> Optional[ObjectSaver]:
        return TorchObjectSaver(value)

    def wrap_return_value(self, value: th.Tensor):
        return value.view(value.size())


MODULE_EXTENSIONS.add(th, TorchModuleExtension())



from abc import ABC
from dataclasses import dataclass, field
from typing import Optional, Dict, TypeVar

from mamo.internal.cached_values import ExternallyCachedFilePath, CachedValue
from mamo.internal.fingerprints import FingerprintDigest, FingerprintDigestRepr, MAX_FINGERPRINT_VALUE_LENGTH
from mamo.internal.reflection import get_module_name


T = TypeVar("T")

MISSING_DIGEST = object()


class ObjectSaver(ABC):
    def __init__(self, value):
        self.value = value
        self.digest = MISSING_DIGEST

    def get_estimated_size(self) -> Optional[int]:
        """Returns None if the size couldn't be estimated."""
        raise NotImplementedError()

    def compute_digest_(self):
        """Returns None if the digest couldn't be created."""
        raise NotImplementedError()

    def compute_digest(self):
        if self.digest is MISSING_DIGEST:
            self.digest = self.compute_digest_()

        return self.digest

    def compute_fingerprint(self):
        """Returns None if the fingerprint couldn't be created."""
        digest = self.compute_digest()
        if digest is None:
            # TODO: log?
            return None

        estimated_size = self.get_estimated_size()
        if estimated_size is not None and estimated_size <= MAX_FINGERPRINT_VALUE_LENGTH:
            return FingerprintDigestRepr(digest, repr(self.value))
        return FingerprintDigest(digest)

    def cache_value(self, external_path_builder: Optional[ExternallyCachedFilePath]) -> Optional[CachedValue]:
        """Returns None if the value couldn't be cached."""
        raise NotImplementedError()


@dataclass
class ModuleExtension:
    module_registry: "ModuleRegistry" = None

    def supports(self, value) -> bool:
        raise NotImplementedError()

    def get_object_saver(self, value) -> Optional[ObjectSaver]:
        """Returns an `ObjectSaver` or None if it couldn't be created."""
        raise NotImplementedError()

    def wrap_return_value(self, value):
        """Returns None if the value couldn't be wrapped."""
        raise NotImplementedError()


@dataclass
class ModuleRegistry:
    default_extension: ModuleExtension = None
    store: Dict[str, ModuleExtension] = field(default_factory=dict, init=False)

    def set_default_extension(self, handler: ModuleExtension):
        self.default_extension = handler
        handler.module_registry = self

    def add(self, module, handler: ModuleExtension):
        self.store[module.__name__] = handler
        handler.module_registry = self

    def get(self, value) -> ModuleExtension:
        module_name = get_module_name(value)
        extension = self.store.get(module_name)
        return extension

    def get_object_saver(self, value) -> ObjectSaver:
        extension = self.get(value)
        object_saver = None
        if extension is not None and extension.supports(value):
            object_saver = extension.get_object_saver(value)

        if object_saver is None:
            object_saver = self.default_extension.get_object_saver(value)
        return object_saver

    def wrap_return_value(self, value):
        # We cannot really do anything about None sadly.
        # Why?
        # Because we usually test for None using 'is\is not None'
        # and having a proxy breaks that.
        # The general rule is that one should not mamo 'None' results

        # TODO: raise or log? (probably throw because it breaks staleness etc?)
        if value is None:
            return None

        extension = self.get(value)
        return_value = None
        if extension is not None and extension.supports(value):
            return_value = extension.wrap_return_value(value)

        if return_value is None:
            return_value = self.default_extension.wrap_return_value(value)

        if return_value is None:
            return_value = value

        return return_value


MODULE_EXTENSIONS = ModuleRegistry()



from mamo.internal import reflection

from mamo.internal.identities import (
    ValueFingerprintIdentity,
    ValueIdentity,
    CellIdentity,
    ValueCallIdentity,
    ValueCellResultIdentity
)
from mamo.internal.providers import IdentityProvider, ValueOracle


class IdentityRegistry(IdentityProvider):
    value_oracle: ValueOracle

    def __init__(self, value_oracle: ValueOracle):
        self.value_oracle = value_oracle

    def identify_call(self, fid, args, kwargs) -> ValueCallIdentity:
        args_vid = tuple(self.value_oracle.identify_value(arg) for arg in args)
        kwargs_vid = frozenset((name, self.value_oracle.identify_value(value)) for name, value in kwargs.items())

        return ValueCallIdentity(fid, args_vid, kwargs_vid)

    def identify_cell_result(self, cell_identity: CellIdentity, key: str) -> ValueCellResultIdentity:
        return ValueCellResultIdentity(cell_identity, key)

    def identify_value(self, value) -> ValueIdentity:
        fingerprint = self.value_oracle.fingerprint_value(value)
        return ValueFingerprintIdentity(reflection.get_type_qualified_name(value), fingerprint)



import os
from typing import Optional

import numpy as np
import hashlib

from mamo.api_support import (
    DBPickledValue,
    ExternallyCachedValue,
    ModuleExtension,
    ObjectSaver,
    ExternallyCachedFilePath,
    CachedValue,
    MODULE_EXTENSIONS,
)

np_types = (np.ndarray, np.record, np.matrix, np.recarray, np.chararray, np.generic, np.memmap)


class NumpyExternallyCachedValue(ExternallyCachedValue):
    def load(self):
        return np.load(self.path, "r")

    @staticmethod
    def save(external_path, value):
        np.save(external_path, value)

        return NumpyExternallyCachedValue(external_path)


class NumpyObjectSaver(ObjectSaver):
    def __init__(self, value):
        super().__init__(value)
        self.value = value

    def get_estimated_size(self) -> Optional[int]:
        return self.value.nbytes

    def compute_digest_(self):
        return hashlib.md5(self.value).digest()

    def cache_value(self, external_path_builder: Optional[ExternallyCachedFilePath]) -> Optional[CachedValue]:
        if external_path_builder is None:
            return DBPickledValue.cache_value(self.value)

        shape_info = "_".join(map(str, self.value.shape))
        external_path = external_path_builder.build(shape_info, "npy")

        return NumpyExternallyCachedValue.save(external_path, self.value)


class NumpyModuleExtension(ModuleExtension):
    def supports(self, value) -> bool:
        return isinstance(value, np_types)

    def get_object_saver(self, value) -> Optional[ObjectSaver]:
        return NumpyObjectSaver(value)

    def wrap_return_value(self, value):
        # We also make value readonly.
        value.setflags(write=False)
        return value.view()


MODULE_EXTENSIONS.add(np, NumpyModuleExtension())



import inspect
from types import CodeType, FunctionType
from typing import Optional, Set, Dict, MutableMapping
from weakref import WeakKeyDictionary

from mamo.internal import reflection
from mamo.internal.fingerprints import (
    FunctionFingerprint,
    DeepFunctionFingerprint,
    CallFingerprint,
    FingerprintDigestRepr,
    CellResultFingerprint,
    CellFingerprint,
    Fingerprint, MAX_FINGERPRINT_VALUE_LENGTH)
from mamo.internal.identities import (
    ValueCallIdentity,
    ComputedValueIdentity,
    ValueIdentityVisitor,
    ValueCellResultIdentity)
from mamo.internal.module_extension import MODULE_EXTENSIONS
from mamo.internal.providers import ValueProvider, FunctionProvider, FingerprintProvider, ValueOracle
from mamo.internal.reflection import FunctionDependencies

# TODO: This can be part of the default module extension! (or its own extension!!)
# We can define a FunctionCall wrapper and pass that through the module system to allow for customization!
from mamo.internal.weakref_utils import WeakKeyIdMap


class FingerprintRegistry(FingerprintProvider):
    # TODO: We only need the FingerprintProvider bit of ValueProvider!
    # Once we move
    value_provider: ValueProvider
    value_oracle: ValueOracle
    function_provider: FunctionProvider

    cache: WeakKeyIdMap[object, Fingerprint]

    # actually a WeakKeyDictionary!
    code_object_deps: MutableMapping[CodeType, FunctionDependencies]
    # If not None, allows for deep function fingerprinting.
    deep_fingerprint_source_prefix: Optional[str]
    deep_fingerprint_stack: Set[CodeType]

    def __init__(
            self,
            deep_fingerprint_source_prefix: Optional[str],
            value_provider: ValueProvider,
            value_oracle: ValueOracle,
            function_provider: FunctionProvider
    ):
        self.cache = WeakKeyIdMap()

        self.code_object_deps = WeakKeyDictionary()

        self.deep_fingerprint_source_prefix = deep_fingerprint_source_prefix
        self.deep_fingerprint_stack = set()

        self.value_provider = value_provider
        self.value_oracle = value_oracle
        self.function_provider = function_provider

    def fingerprint_value(self, value):
        # TODO: do I want to store strings like that?
        if value is None or isinstance(value, (bool, int, float)) or (isinstance(value, str) and len(value) <
                                                                      MAX_FINGERPRINT_VALUE_LENGTH):
            # TODO: special-case strings and summarize them?
            return FingerprintDigestRepr(value, repr(value))

        fingerprint = self.cache.get(value)
        if fingerprint is not None:
            return fingerprint

        # TODO: this is a special case of a computing a digest!?
        if isinstance(value, FunctionType):
            fingerprint = FingerprintDigestRepr(self._get_function_fingerprint(value, allow_deep=False),
                                                reflection.get_func_qualified_name(value))
        else:
            object_saver = MODULE_EXTENSIONS.get_object_saver(value)
            if object_saver is not None:
                fingerprint = object_saver.compute_fingerprint()

        if fingerprint is None:
            # TODO: log?
            raise ValueError(
                f"Cannot fingerprint {value}!"
                " Please either add an extension to support it,"
                " or register it with a name"
            )

        self.cache[value] = fingerprint

        return fingerprint

    def fingerprint_function(self, func):
        # Function fingerprints (when we allow deep fingerprints) cannot be cached
        return self._get_function_fingerprint(func, allow_deep=True)

    def fingerprint_call(self, func: FunctionType, args, kwargs: Dict):
        # Call fingerprints (when we allow deep fingerprints) cannot be cached
        func_fingerprint = self._get_function_fingerprint(func)
        args_fingerprints = tuple(self.value_oracle.fingerprint_value(arg) for arg in args)
        kwargs_fingerprints = frozenset((name, self.value_oracle.fingerprint_value(arg)) for name, arg in kwargs.items())

        return CallFingerprint(func_fingerprint, args_fingerprints, kwargs_fingerprints)

    def fingerprint_cell(self, cell_function: FunctionType) -> CellFingerprint:
        cell_code_fingerprint = self._get_deep_fingerprint(cell_function.__code__, cell_function.__globals__)

        func_deps = reflection.get_func_deps(cell_function)
        global_loads = func_deps.global_loads
        global_stores = func_deps.global_stores

        resolved_globals_loads = reflection.resolve_qualified_names(global_loads, cell_function.__globals__)
        globals_load_fingerprint = frozenset(
            (name, (self.value_oracle.identify_value(value), self.value_oracle.fingerprint_value(value)))
            for name, value in resolved_globals_loads.items()
        )
        cell_fingerprint = CellFingerprint(cell_code_fingerprint, globals_load_fingerprint, frozenset(global_stores))
        return cell_fingerprint

    def fingerprint_cell_result(self, cell_fingerprint: CellFingerprint, key: str):
        return CellResultFingerprint(cell_fingerprint, key)

    def fingerprint_computed_value(self, vid: ComputedValueIdentity):
        # TODO: move back to main.py:is_stale_vid!
        outer_self = self

        class Visitor(ValueIdentityVisitor):
            def visit_call(self, vid: ValueCallIdentity):
                func_fingerprint = outer_self.fingerprint_function(
                    outer_self.function_provider.resolve_function(vid.fid)
                )
                arg_fingerprints = [
                    outer_self.value_provider.resolve_fingerprint(arg_vid) for arg_vid in vid.args_vid
                ]
                kwarg_fingerprints = [
                    (name, outer_self.value_provider.resolve_fingerprint(arg_vid))
                    for name, arg_vid in vid.kwargs_vid
                ]
                return CallFingerprint(func_fingerprint, tuple(arg_fingerprints), frozenset(kwarg_fingerprints))

            def visit_cell_result(self, vid: ValueCellResultIdentity):
                cell_function = outer_self.function_provider.resolve_function(vid.cell)
                return outer_self.fingerprint_cell_result(outer_self.fingerprint_cell(cell_function), vid.key)

        return Visitor().visit(vid)

    def _get_code_object_deps(self, code_object) -> FunctionDependencies:
        code_object_deps = self.code_object_deps.get(code_object)
        if code_object_deps is None:
            code_object_deps = reflection.get_func_deps(code_object)
            self.code_object_deps[code_object] = code_object_deps
        return code_object_deps

    def _get_deep_fingerprint(self, code_object, namespace):
        if code_object in self.deep_fingerprint_stack:
            return FunctionFingerprint(reflection.get_code_object_fingerprint(code_object))
        self.deep_fingerprint_stack.add(code_object)
        try:
            func_deps = self._get_code_object_deps(code_object)

            resolved_funcs = reflection.resolve_qualified_names(func_deps.func_calls, namespace)

            # TODO: this does not seem to resolve builtins!!?!?! debug

            global_funcs = {
                qn: self._get_function_fingerprint(resolved_func, allow_deep=True)
                for qn, resolved_func in resolved_funcs.items() if resolved_func
            }

            return DeepFunctionFingerprint(
                reflection.get_code_object_fingerprint(code_object), frozenset(global_funcs.items())
            )
        finally:
            self.deep_fingerprint_stack.remove(code_object)

    def _get_function_fingerprint(self, callee, allow_deep=True) -> Optional[FunctionFingerprint]:
        # TODO: necessary?
        if callee is None:
            return None
        elif reflection.is_func_builtin(callee):
            func_fingerprint = FunctionFingerprint(callee.__qualname__)
        else:
            # TODO: more tests? code review?
            if isinstance(callee, FunctionType):
                func = callee
            else:
                assert callable(callee), callee
                # Needs tests!
                if isinstance(callee, type):
                    sub_callable = callee.__init__
                elif hasattr(callee, '__func__'):
                    sub_callable = callee.__func__
                elif hasattr(callee, '__call__'):
                    sub_callable = callee.__call__
                else:
                    # log and fail softly?
                    raise NotImplementedError(f'Missing support for fingerprinting callable {callee}!')
                return self._get_function_fingerprint(sub_callable, allow_deep=allow_deep)

            if not hasattr(callee, '__code__') or not hasattr(callee, '__globals__'):
                import types
                isbuiltin = inspect.isbuiltin(callee)

            # Unwrap special functions.
            if hasattr(callee, "mamo_unwrapped_func"):
                callee = callee.mamo_unwrapped_func

            if allow_deep and reflection.is_func_local(callee, self.deep_fingerprint_source_prefix):
                func_fingerprint = self._get_deep_fingerprint(callee.__code__, callee.__globals__)
            else:
                func_fingerprint = FunctionFingerprint(reflection.get_func_fingerprint(callee))

        return func_fingerprint



from mamo.internal.weakref_utils import WeakIdSet, supports_weakrefs


class StalenessRegistry:
    # We need to keep track of unlinked values to be able to tell that they are stale now!
    stale_values: WeakIdSet[object]

    def __init__(self):
        super().__init__()
        self.stale_values = WeakIdSet()

    def mark_stale(self, value):
        if supports_weakrefs(value):
            self.stale_values.add(value)

    def mark_used(self, value):
        if supports_weakrefs(value):
            self.stale_values.discard(value)

    def is_stale(self, value):
        return value in self.stale_values



# Inspired by https://stackoverflow.com/a/30024601/854731
from contextlib import AbstractContextManager
from timeit import default_timer


class StopwatchContext(AbstractContextManager):
    """Usage:

    ```
    with StopwatchContext() as stopwatch:
        ...

    print(stopwatch.elapsed_time)
    ```
    """

    def __init__(self):
        self.start_time = None
        self.end_time = None

    @property
    def elapsed_time(self):
        return self.end_time - self.start_time

    def __enter__(self):
        self.start_time = default_timer()
        return super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        self.end_time = default_timer()



import ast
import dataclasses
from typing import Optional, List

from functools import wraps

from mamo.internal.fingerprint_registry import FingerprintRegistry
from mamo.internal.fingerprints import Fingerprint, CellResultFingerprint, ResultFingerprint
from mamo.internal.identities import (
    ValueIdentity,
    value_name_identity,
    ComputedValueIdentity,
    ValueCallIdentity,
    ValueCellResultIdentity)
from mamo.internal.identity_registry import IdentityRegistry
from mamo.internal.function_registry import FunctionRegistry
from mamo.internal.module_extension import MODULE_EXTENSIONS
from mamo.internal.result_metadata import ResultMetadata
from mamo.internal.result_registry import ResultRegistry
from mamo.internal.stopwatch_context import StopwatchContext
from mamo.internal.value_provider_mediator import ValueProviderMediator
from mamo.internal.value_registries import ValueRegistry
from mamo.internal.staleness_registry import StalenessRegistry
from mamo.internal.persisted_store import PersistedStore

from mamo.internal import default_module_extension

# Install the default module extension.
MODULE_EXTENSIONS.set_default_extension(default_module_extension.DefaultModuleExtension())


class ReExecutionPolicy:
    def __call__(
            self,
            mamo: "Mamo",
            vid: ComputedValueIdentity,
            fingerprint: ResultFingerprint,
            stored_fingerprint: Optional[ResultFingerprint],
    ):
        return True


def execute_decision_only_missing(
        mamo: "Mamo", vid: ComputedValueIdentity, fingerprint: Fingerprint, stored_fingerprint: Optional[Fingerprint]
):
    return False


def execute_decision_stale(max_depth):
    def decider(
            mamo: "Mamo", vid: ComputedValueIdentity, fingerprint: ResultFingerprint,
            stored_fingerprint: Optional[ResultFingerprint]
    ):
        if fingerprint != stored_fingerprint:
            return True
        return mamo.is_stale_vid(vid, depth=max_depth)

    return decider


class Mamo:
    fingerprint_registry: FingerprintRegistry
    identity_registry: IdentityRegistry
    staleness_registry: StalenessRegistry
    function_registry: FunctionRegistry

    value_provider_mediator: ValueProviderMediator
    external_value_registry: ValueRegistry
    result_registry: ResultRegistry

    persisted_store: PersistedStore

    re_execution_policy: ReExecutionPolicy

    _call_duration_stack: List
    _nomamo_call_duration_stack: List

    def __init__(self, persisted_store, deep_fingerprint_source_prefix: Optional[str],
                 re_execution_policy: Optional[ReExecutionPolicy]):
        self.persisted_store = persisted_store

        self.value_provider_mediator = ValueProviderMediator()

        self.staleness_registry = StalenessRegistry()
        self.external_value_registry = ValueRegistry(self.staleness_registry)
        self.result_registry = ResultRegistry(self.staleness_registry, persisted_store)

        self.function_registry = FunctionRegistry()
        self.fingerprint_registry = FingerprintRegistry(deep_fingerprint_source_prefix, self.value_provider_mediator,
                                                        self.value_provider_mediator, self.function_registry)
        self.identity_registry = IdentityRegistry(self.value_provider_mediator)

        self.value_provider_mediator.init(self.identity_registry, self.fingerprint_registry, self.result_registry,
                                          self.external_value_registry)

        self.re_execution_policy = re_execution_policy or execute_decision_stale(-1)

        self._call_duration_stack = [0.0]
        self._nomamo_call_duration_stack = [0.0]

    def swap_persisted_store(self, new_persisted_store):
        self.persisted_store.close()
        self.persisted_store = new_persisted_store
        self.result_registry.persisted_store = new_persisted_store

    @property
    def deep_fingerprint_source_prefix(self):
        return self.fingerprint_registry.deep_fingerprint_source_prefix

    # TODO: remove this property again (only used by tests!)

    @deep_fingerprint_source_prefix.setter
    def deep_fingerprint_source_prefix(self, value):
        self.fingerprint_registry.deep_fingerprint_source_prefix = value

    def _get_value(self, vid: ValueIdentity):
        return self.value_provider_mediator.resolve_value(vid)

    def _get_vid(self, value: object):
        return self.value_provider_mediator.identify_value(value)

    def get_value_identities(self, persisted=False):
        # TODO: optimize to always create a new set?
        vids = self.value_provider_mediator.get_vids()
        if persisted:
            vids.update(self.persisted_store.get_vids())
        return vids

    def flush_cache(self):
        self.result_registry.flush()

    # TODO: add tests
    def flush_value(self, value):
        self.result_registry.flush_value(value)

    def is_stale_call(self, func, args, kwargs, *, depth=-1):
        fid = self.function_registry.identify_function(func)
        vid = self.identity_registry.identify_call(fid, args, kwargs)

        return self.is_stale_vid(vid, depth=depth)

    def is_stale(self, value, *, depth=-1):
        if self.staleness_registry.is_stale(value):
            # More interesting result type?!
            # Value has become stale!'
            return True

        vid = self._get_vid(value)
        if vid is None:
            # TODO: throw?
            print('Vid not found!')
            return True

        return self.is_stale_vid(vid, depth=depth)

    def is_stale_vid(self, vid: Optional[ValueIdentity], *, depth):
        if vid is None:
            return False
        if not isinstance(vid, ComputedValueIdentity):
            return False

        fingerprint = self.fingerprint_registry.fingerprint_computed_value(vid)
        stored_fingerprint = self.value_provider_mediator.resolve_fingerprint(vid)

        if fingerprint != stored_fingerprint:
            print(f'{vid} is stale!')
            print(f'{fingerprint}\nvs\n{stored_fingerprint}')
            return True

        if depth == 0:
            return False

        if isinstance(vid, ValueCallIdentity):
            return any(self.is_stale_vid(arg_vid, depth=depth - 1) for arg_vid in vid.args_vid) or any(
                self.is_stale_vid(arg_vid, depth=depth - 1) for name, arg_vid in vid.kwargs_vid
            )
        elif isinstance(vid, ValueCellResultIdentity):
            assert isinstance(fingerprint, CellResultFingerprint)
            return any(
                self.is_stale_vid(input_vid, depth=depth - 1)
                for name, (input_vid, input_fingerprint) in fingerprint.cell.globals_load
            )

    def is_cached_call(self, func, args, kwargs):
        fid = self.function_registry.identify_function(func)
        vid = self.identity_registry.identify_call(fid, args, kwargs)

        return self.value_provider_mediator.has_vid(vid)

    def forget_call(self, func, args, kwargs):
        fid = self.function_registry.identify_function(func)
        vid = self.identity_registry.identify_call(fid, args, kwargs)

        self.value_provider_mediator.remove_vid(vid)

    def forget(self, value):
        if not self.value_provider_mediator.has_value(value):
            # TODO: throw or log
            return
        self.value_provider_mediator.remove_value(value)

    def get_metadata_call(self, func, args, kwargs) -> Optional[ResultMetadata]:
        fid = self.function_registry.identify_function(func)
        vid = self.identity_registry.identify_call(fid, args, kwargs)

        metadata = self.persisted_store.get_result_metadata(vid)
        if not metadata:
            return None
        return dataclasses.replace(metadata)

    def get_metadata(self, value) -> Optional[ResultMetadata]:
        vid = self.value_provider_mediator.identify_value(value)
        metadata = self.persisted_store.get_result_metadata(vid)

        if not metadata:
            return None
        return dataclasses.replace(metadata)

    def _shall_execute(self, vid: ComputedValueIdentity, fingerprint: ResultFingerprint):
        # TODO: could directly ask persisted_cache
        stored_fingerprint = mamo.result_registry.resolve_fingerprint(vid)
        return stored_fingerprint is None or self.re_execution_policy(self, vid, fingerprint, stored_fingerprint)

    @staticmethod
    def wrap_function(func):
        @wraps(func)
        def wrapped_func(*args, **kwargs):
            with StopwatchContext() as total_stopwatch:
                nonlocal fid

                # If mamo was not initialized before, we might still have to set fid.
                if fid is None:
                    # Just initialize it with defaults.
                    if mamo is None:
                        # TODO: maybe log?
                        init_mamo()

                    fid = mamo.function_registry.identify_function(func)

                vid = mamo.identity_registry.identify_call(fid, args, kwargs)

                call_fingerprint = mamo.fingerprint_registry.fingerprint_call(func, args, kwargs)
                result_metadata = mamo.persisted_store.get_result_metadata(vid)

                if mamo._shall_execute(vid, call_fingerprint):
                    mamo._call_duration_stack.append(0.)
                    mamo._nomamo_call_duration_stack.append(0.)
                    with StopwatchContext() as call_stopwatch:
                        result = func(*args, **kwargs)
                    wrapped_result = MODULE_EXTENSIONS.wrap_return_value(result)
                    mamo.value_provider_mediator.add(vid, wrapped_result, call_fingerprint)

                    result_metadata = mamo.persisted_store.get_result_metadata(vid)
                    result_metadata.call_duration = call_stopwatch.elapsed_time
                    result_metadata.subcall_duration = mamo._call_duration_stack.pop()
                    result_metadata.estimated_nomamo_call_duration = result_metadata.call_duration - result_metadata.subcall_duration + mamo._nomamo_call_duration_stack.pop()
                else:
                    wrapped_result = mamo._get_value(vid)
                    if wrapped_result is None:
                        # log?
                        raise RuntimeError(f"Couldn't find cached result for {vid}!")

                    result_metadata.num_cache_hits += 1

            result_metadata.total_durations += total_stopwatch.elapsed_time
            mamo._call_duration_stack[-1] += total_stopwatch.elapsed_time
            mamo._nomamo_call_duration_stack[-1] += result_metadata.estimated_nomamo_call_duration
            return wrapped_result

        wrapped_func.mamo_unwrapped_func = func
        wrapped_func.is_stale = lambda *args, **kwargs: mamo.is_stale_call(func, args, kwargs)
        wrapped_func.is_cached = lambda *args, **kwargs: mamo.is_cached_call(func, args, kwargs)
        wrapped_func.forget = lambda *args, **kwargs: mamo.forget_call(func, args, kwargs)
        wrapped_func.get_metadata = lambda *args, **kwargs: mamo.get_metadata_call(func, args, kwargs)
        wrapped_func.get_tag_name = lambda *args, **kwargs: mamo.get_tag_name_call(func, args, kwargs)

        # This method is a static method, so that mamo does not need to be initialized.
        fid = None
        if mamo is not None:
            fid = mamo.function_registry.identify_function(func)

        return wrapped_func

    def run_cell(self, name: Optional[str], cell_code: str, user_ns: dict):
        # TODO: wrap in a function and execute, so we need explicit globals for stores?
        function_module = ast.parse("def cell_function():\n  pass")
        cell_module = ast.parse(cell_code)
        function_module.body[0].body = cell_module.body
        compiled_function = compile(function_module, "<code>", "exec")

        local_ns = {}
        exec(compiled_function, user_ns, local_ns)
        cell_function = local_ns["cell_function"]

        cell_id = self.function_registry.identify_cell(name, cell_function)
        cell_fingerprint = self.fingerprint_registry.fingerprint_cell(cell_function)

        outputs = cell_fingerprint.outputs

        result_vids = {name: self.identity_registry.identify_cell_result(cell_id, name) for name in outputs}
        result_fingerprints = {
            name: self.fingerprint_registry.fingerprint_cell_result(cell_fingerprint, name) for name in outputs
        }

        # TODO: this adds some staleness overhead but not sure how to handle composites atm.
        if any(mamo._shall_execute(result_vids[name], result_fingerprints[name]) for name in outputs):
            cell_function()

            # Retrieve stores.
            wrapped_results = {name: MODULE_EXTENSIONS.wrap_return_value(user_ns[name]) for name in outputs}
            user_ns.update(wrapped_results)

            for name in outputs:
                mamo.value_provider_mediator.add(result_vids[name],
                                                  user_ns[name], result_fingerprints[name])
        else:
            for name in outputs:
                vid = result_vids[name]
                cached_result = mamo._get_value(vid)
                if cached_result is None:
                    # log?
                    raise RuntimeError(f"Couldn't find cached result for {vid}!")

                user_ns[name] = cached_result

    def tag(self, tag_name: str, value: Optional[object]):
        vid = None
        # Value should exist in the cache.
        if value is not None:
            vid = self.value_provider_mediator.identify_value(value)
            if vid is None:
                raise ValueError("Value has not been registered previously!")

        self.persisted_store.tag(tag_name, vid)

    def get_tag_name_call(self, func, args, kwargs) -> Optional[str]:
        fid = self.function_registry.identify_function(func)
        vid = self.identity_registry.identify_call(fid, args, kwargs)

        return self.persisted_store.get_tag_name(vid)

    def get_tag_name(self, value) -> Optional[str]:
        vid = self._get_vid(value)
        if not vid:
            return None
        return self.persisted_store.get_tag_name(vid)

    def get_tag_value(self, tag_name):
        # TODO: might have to expose has_tag etc?
        vid = self.persisted_store.get_tag_vid(tag_name)
        if vid is None:
            # TODO: log instead
            # raise ValueError(f"{tag_name} has not been registered previously!")
            return None
        value = self.value_provider_mediator.resolve_value(vid)
        if value is None:
            # TODO: log instead!
            # raise ValueError(f"{vid} for {tag_name} is not available anymore!")
            return None
        return value

    def register_external_value(self, unique_name, value):
        vid = value_name_identity(unique_name)
        if value is None:
            self.value_provider_mediator.remove_vid(vid)
        else:
            self.value_provider_mediator.add(vid, value, vid.fingerprint)
        # TODO: add a test!

        return value

    def get_external_value(self, unique_name):
        vid = value_name_identity(unique_name)
        return self.value_provider_mediator.resolve_value(vid)

    # noinspection PyTypeChecker
    def testing_close(self):
        self.persisted_store.close()
        self.identity_registry = None
        self.fingerprint_registry = None
        self.value_provider_mediator = None
        import gc
        gc.collect()


mamo: Optional[Mamo] = None


def init_mamo(
        memory_only=True,
        path: Optional[str] = None,
        externally_cached_path: Optional[str] = None,
        # By default, we don't use deep fingerprints except in the main module/jupyter notebooks.
        deep_fingerprint_source_prefix: Optional[str] = None,
        re_execution_policy: Optional[ReExecutionPolicy] = None
):
    global mamo
    assert mamo is None

    new_persisted_store = (
        PersistedStore.from_memory()
        if memory_only
        else PersistedStore.from_file(path, externally_cached_path)
    )
    mamo = Mamo(new_persisted_store, deep_fingerprint_source_prefix, re_execution_policy)


# TODO: add tests!
# TODO: This is a hack because it breaks metadata!
def swap_storage(
    memory_only=True,
    path: Optional[str] = None,
    externally_cached_path: Optional[str] = None
):
    assert mamo is not None

    new_persisted_store = (
        PersistedStore.from_memory()
        if memory_only
        else PersistedStore.from_file(path, externally_cached_path)
    )

    mamo.swap_persisted_store(new_persisted_store)



from typing import Optional, Set

from mamo.internal.fingerprints import Fingerprint
from mamo.internal.identities import ValueIdentity, ComputedValueIdentity, ValueFingerprintIdentity
from mamo.internal.providers import ValueProvider, IdentityProvider, FingerprintProvider


class ValueProviderMediator(ValueProvider):
    identity_provider: IdentityProvider
    fingerprint_provider: FingerprintProvider
    result_provider: ValueProvider
    external_value_provider: ValueProvider

    def init(self, identity_provider: IdentityProvider,
             fingerprint_provider: FingerprintProvider,
             result_provider: ValueProvider,
             external_value_provider: ValueProvider):
        self.fingerprint_provider = fingerprint_provider
        self.identity_provider = identity_provider
        self.result_provider = result_provider
        self.external_value_provider = external_value_provider

    def identify_value(self, value) -> ValueIdentity:
        return (self.external_value_provider.identify_value(value) or self.result_provider.identify_value(
            value) or self.identity_provider.identify_value(value))

    def fingerprint_value(self, value) -> Fingerprint:
        return (self.external_value_provider.fingerprint_value(value) or self.result_provider.fingerprint_value(
            value) or self.fingerprint_provider.fingerprint_value(value))

    def resolve_value(self, vid: ValueIdentity):
        if isinstance(vid, ComputedValueIdentity):
            return self.result_provider.resolve_value(vid)
        return self.external_value_provider.resolve_value(vid)

    def resolve_fingerprint(self, vid: ValueIdentity):
        if isinstance(vid, ValueFingerprintIdentity):
            return vid.fingerprint
        elif isinstance(vid, ComputedValueIdentity):
            return self.result_provider.resolve_fingerprint(vid)
        return self.external_value_provider.resolve_fingerprint(vid)

    def add(self, vid: ValueIdentity, value, fingerprint: Fingerprint):
        if self.has_value(value):
            existing_vid = self.identify_value(value)
            assert existing_vid is not None
            if existing_vid != vid:
                raise AttributeError(
                    f"{vid} has same value as {existing_vid}!"
                    'We follow an "each computation, different result" policy.'
                    "This makes tracking possible."
                )

        # Vids are compartmentalized by value registry and thus we don't need any
        # additional error checking here.

        if isinstance(vid, ComputedValueIdentity):
            return self.result_provider.add(vid, value, fingerprint)
        else:
            return self.external_value_provider.add(vid, value, fingerprint)

    def remove_vid(self, vid: ValueIdentity):
        if isinstance(vid, ComputedValueIdentity):
            return self.result_provider.remove_vid(vid)
        else:
            return self.external_value_provider.remove_vid(vid)

    def remove_value(self, value):
        self.result_provider.remove_value(value)
        self.external_value_provider.remove_value(value)

    def has_vid(self, vid: ValueIdentity):
        return self.external_value_provider.has_vid(vid) or self.result_provider.has_vid(vid)

    def has_value(self, value):
        return self.external_value_provider.has_value(value) or self.result_provider.has_value(value)

    def get_vids(self) -> Set[ValueIdentity]:
        vids = set()
        vids.update(self.external_value_provider.get_vids())
        vids.update(self.result_provider.get_vids())
        return vids



from abc import ABC
from types import FunctionType
from typing import NoReturn, Set

from mamo.internal.fingerprints import Fingerprint
from mamo.internal.identities import ValueIdentity, FunctionIdentity, CellIdentity


class FunctionProvider:
    def identify_function(self, function: FunctionType) -> FunctionIdentity:
        raise NotImplementedError()

    def identify_cell(self, name: str, cell_function: FunctionType) -> CellIdentity:
        raise NotImplementedError()

    def resolve_function(self, fid: FunctionIdentity) -> FunctionType:
        raise NotImplementedError()


class IdentityProvider:
    def identify_value(self, value) -> ValueIdentity:
        raise NotImplementedError()


class FingerprintProvider:
    def fingerprint_value(self, value):
        raise NotImplementedError()


class ValueOracle(IdentityProvider, FingerprintProvider, ABC):
    pass


class ValueProvider(ValueOracle, ABC):
    def resolve_value(self, vid: ValueIdentity) -> object:
        raise NotImplementedError()

    def resolve_fingerprint(self, vid: ValueIdentity) -> Fingerprint:
        raise NotImplementedError()

    def add(self, vid: ValueIdentity, value, fingerprint: Fingerprint) -> NoReturn:
        raise NotImplementedError()

    def remove_vid(self, vid: ValueIdentity) -> NoReturn:
        raise NotImplementedError()

    def remove_value(self, value: object) -> NoReturn:
        raise NotImplementedError()

    def has_vid(self, vid: ValueIdentity) -> bool:
        raise NotImplementedError()

    def has_value(self, value) -> bool:
        raise NotImplementedError()

    def get_vids(self) -> Set[ValueIdentity]:
        raise NotImplementedError()



from dataclasses import dataclass


@dataclass
class ResultMetadata:
    result_size: int
    stored_size: int
    save_duration: float

    total_load_durations: float = 0
    num_loads: int = 0

    num_cache_hits: int = 0
    total_durations: float = 0

    call_duration: float = 0
    subcall_duration: float = 0

    estimated_nomamo_call_duration: float = 0

    @property
    def avg_total_duration(self):
        return self.total_durations / self.num_calls

    @property
    def avg_overhead_duration(self):
        return (self.total_durations - self.total_load_durations - self.call_duration) / self.num_calls

    @property
    def avg_load_duration(self):
        return self.total_load_durations / self.num_loads

    @property
    def estimated_saved_time(self):
        return (self.estimated_nomamo_call_duration - self.avg_total_duration) * self.num_calls

    @property
    def num_calls(self):
        return self.num_cache_hits + 1



from types import FunctionType
from typing import Dict

from mamo.internal import reflection
from mamo.internal.identities import FunctionIdentity, CellIdentity
from mamo.internal.providers import FunctionProvider


class FunctionRegistry(FunctionProvider):
    fid_to_func: Dict[FunctionIdentity, FunctionType]

    def __init__(self):
        super().__init__()
        self.fid_to_func = {}

    def identify_function(self, func) -> FunctionIdentity:
        fid = FunctionIdentity(reflection.get_func_qualified_name(func))
        self.fid_to_func[fid] = func
        return fid

    def identify_cell(self, name: str, cell_function: FunctionType) -> CellIdentity:
        fid = CellIdentity(name)
        self.fid_to_func[fid] = cell_function
        return fid

    def resolve_function(self, fid):
        return self.fid_to_func.get(fid)



from dataclasses import dataclass
from typing import MutableMapping, Generic, TypeVar, Optional

from persistent.mapping import PersistentMapping
from persistent import Persistent


KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.


class Bimap(Generic[KT, VT]):
    def update(self, key: KT, value: Optional[VT]):
        if key is None and value is None:
            raise ValueError("key and value both None!")

        if key is None:
            self.del_value(value)
        elif value is None:
            self.del_key(key)
        else:
            existing_key = self.get_key(value)
            existing_value = self.get_value(key)

            if key is existing_key and value is existing_value:
                return

            if existing_key is not None:
                self._del_key(existing_key)

            if existing_value is not None:
                self._del_value(existing_value)

            self.put_key_value(key, value)

    def put_key_value(self, key: KT, value: VT):
        if key is None or value is None:
            if key is not None:
                raise ValueError("Value is None! Use update instead!")
            if value is not None:
                raise ValueError("Key is None! Use update instead!")
            raise ValueError("Key and value are None! Use update instead!")

        self._put_key_value(key, value)

    def del_key(self, key: KT):
        existing_value = self.get_value(key)
        if existing_value is None:
            return
        self._del_key(key)
        self._del_value(existing_value)

    def del_value(self, value: VT):
        existing_key = self.get_key(value)
        if existing_key is None:
            return
        self._del_key(existing_key)
        self._del_value(value)

    def length(self):
        raise NotImplementedError()

    def get_key(self, value: VT) -> KT:
        raise NotImplementedError()

    def get_value(self, key: KT) -> VT:
        raise NotImplementedError()

    def _del_key(self, key: KT):
        raise NotImplementedError()

    def _del_value(self, value: VT):
        raise NotImplementedError()

    def _put_key_value(self, key: KT, value: VT):
        raise NotImplementedError()

    # TODO: replace this and below by two methods: keys() and values() which both return sets!
    def __contains__(self, key: KT) -> bool:
        raise NotImplementedError()

    def has_value(self, value: VT) -> bool:
        raise NotImplementedError()


@dataclass
class MappingBimap(Bimap[KT, VT]):
    key_value: MutableMapping[KT, VT]
    value_key: MutableMapping[VT, KT]

    def length(self):
        return len(self.key_value)

    def get_key(self, value):
        return self.value_key.get(value)

    def get_value(self, key):
        return self.key_value.get(key)

    def _del_key(self, key):
        del self.key_value[key]

    def _del_value(self, value):
        del self.value_key[value]

    def _put_key_value(self, key, value):
        self.key_value[key] = value
        self.value_key[value] = key

    def __contains__(self, key):
        return key in self.key_value

    def has_value(self, value):
        return value in self.value_key

    def get_keys(self):
        return self.key_value.keys()


class DictBimap(MappingBimap[KT, VT]):
    def __init__(self):
        super().__init__({}, {})


class PersistentBimap(MappingBimap[KT, VT], Persistent):
    def __init__(self):
        super().__init__(PersistentMapping(), PersistentMapping())



import pickle
from dataclasses import dataclass
from typing import Optional

from mamo.internal.cached_values import CachedValue
from mamo.internal.module_extension import ObjectSaver


@dataclass(frozen=True)
class DBPickledValue(CachedValue):
    """A value that is cached in the database."""

    data: bytes

    @staticmethod
    def cache_value(value) -> Optional[CachedValue]:
        data = try_pickle(value)
        if data is None:
            return None
        return DBPickledValue(data)

    def load(self):
        return try_unpickle(self.data)

    def get_stored_size(self):
        return len(self.data)


def try_pickle(value):
    try:
        return pickle.dumps(value)
    except pickle.PickleError as err:
        # TODO: log err
        print(err)
        return None


def try_unpickle(data: bytes):
    try:
        return pickle.loads(data)
    except pickle.PickleError as err:
        # TODO: log err
        print(err)
        return None



# noinspection PyUnresolvedReferences
from mamo.internal.cached_values import CachedValue, ExternallyCachedFilePath, ExternallyCachedValue
from mamo.internal.db_stored_value import DBPickledValue

# noinspection PyUnresolvedReferences
from mamo.internal.module_extension import ModuleExtension, ObjectSaver, MODULE_EXTENSIONS



import os
import pickle
from dataclasses import dataclass
from typing import Optional, Dict

from ZODB import DB
from ZODB.FileStorage.FileStorage import FileStorage
from persistent import Persistent
from persistent.mapping import PersistentMapping
from transaction import TransactionManager

from mamo.internal.bimap import PersistentBimap
from mamo.internal.cached_values import CachedValue, ExternallyCachedFilePath, ExternallyCachedValue
from mamo.internal.fingerprints import Fingerprint
from mamo.internal.identities import ValueIdentity
from mamo.internal.module_extension import MODULE_EXTENSIONS
from mamo.internal.result_metadata import ResultMetadata
from mamo.internal.stopwatch_context import StopwatchContext
from mamo.internal.weakref_utils import ObjectProxy

MAX_DB_CACHED_VALUE_SIZE = 1024


@dataclass
class MamoPersistedCacheStorage(Persistent):
    external_cache_id: int
    vid_to_cached_value: Dict[ValueIdentity, CachedValue]
    vid_to_fingerprint: Dict[ValueIdentity, Fingerprint]
    vid_to_result_metadata: Dict[ValueIdentity, ResultMetadata]
    tag_to_vid: PersistentBimap[str, ValueIdentity]

    def __init__(self):
        self.vid_to_cached_value = PersistentMapping()
        self.vid_to_fingerprint = PersistentMapping()
        self.vid_to_result_metadata = PersistentMapping()
        self.tag_to_vid = PersistentBimap()
        self.external_cache_id = 0

    def get_new_external_id(self):
        drawn_external_cache_id = self.external_cache_id
        self.external_cache_id += 1
        return f"{drawn_external_cache_id:010}"


@dataclass
class CacheOperationResult:
    cached_value: CachedValue
    result_size: int
    stored_size: int
    save_duration: float


# TODO: to repr method
@dataclass
class PersistedStore:
    storage: MamoPersistedCacheStorage
    transaction_manager: TransactionManager
    db: DB
    path: str
    externally_cached_path: str

    @staticmethod
    def from_memory():
        db = DB(None, large_record_size=64*(1 << 20))
        return PersistedStore(db, None, None)

    @staticmethod
    def from_file(path: Optional[str] = None, externally_cached_path: Optional[str] = None):
        if path is None:
            path = "./"
        if externally_cached_path is None:
            externally_cached_path = path

        # Create directories if they haven't been created yet.
        os.makedirs(path, exist_ok=True)
        os.makedirs(externally_cached_path, exist_ok=True)

        # TODO: log the paths?
        # TODO: in general, make properties available for quering in the console/Jupyter?

        db = DB(FileStorage(os.path.join(path, "mamo_store")))
        return PersistedStore(db, path, externally_cached_path)

    def __init__(self, db: DB, path: Optional[str], externally_cached_path: Optional[str]):
        self.db = db
        self.path = os.path.abspath(path) if path else path
        self.externally_cached_path = (
            os.path.abspath(externally_cached_path) if externally_cached_path else externally_cached_path
        )
        self.transaction_manager = TransactionManager()

        connection = db.open(self.transaction_manager)
        root = connection.root

        if not hasattr(root, "storage"):
            with self.transaction_manager:
                root.storage = MamoPersistedCacheStorage()

        self.storage = root.storage

    def close(self):
        self.db.close()
        self.transaction_manager.clearSynchs()

    def get_new_external_id(self):
        return self.storage.get_new_external_id()

    def try_create_cached_value(
            self, vid: ValueIdentity, value: object
    ) -> Optional[CacheOperationResult]:
        with StopwatchContext() as stopwatch:
            assert value is not None
            object_saver = MODULE_EXTENSIONS.get_object_saver(value)
            if not object_saver:
                # TODO: log?
                return None

            estimated_size = object_saver.get_estimated_size()
            if estimated_size is None:
                # TODO: log?
                return None

            external_path_builder = None
            # If we exceed a reasonable size, we don't store the result in the DB.
            # However, if we are memory-only, we don't cache in external files.
            if estimated_size > MAX_DB_CACHED_VALUE_SIZE and self.externally_cached_path is not None:
                external_path_builder = ExternallyCachedFilePath(
                    self.externally_cached_path, self.get_new_external_id(), vid.get_external_info()
                )

            cached_value = object_saver.cache_value(external_path_builder)
            if not cached_value:
                # TODO: log?
                return None

            stored_size = cached_value.get_stored_size()

        return CacheOperationResult(cached_value=cached_value, result_size=estimated_size, stored_size=stored_size,
                                    save_duration=stopwatch.elapsed_time)

    def add(self, vid: ValueIdentity, value: object, fingerprint: Fingerprint):
        assert value is not None

        with self.transaction_manager:
            existing_cached_value = self.storage.vid_to_cached_value.get(vid)
            if existing_cached_value:
                # assert isinstance(existing_cached_value, CachedValue)
                # TODO: add test cases for unlinking!!!
                existing_cached_value.unlink()

            # TODO: logic to decide whether to store the value at all or not depending
            # on computational budget.

            # Unwrap object proxy
            if isinstance(value, ObjectProxy):
                value = value.__subject__

            result = self.try_create_cached_value(vid, value)
            if result.cached_value:
                self.storage.vid_to_cached_value[vid] = result.cached_value
                self.storage.vid_to_fingerprint[vid] = fingerprint

                result_metadata = ResultMetadata(result_size=result.result_size,
                                                 stored_size=result.stored_size,
                                                 save_duration=result.save_duration)
                self.storage.vid_to_result_metadata[vid] = result_metadata
            else:
                # TODO: log? result is None means caching has failed!
                if existing_cached_value:
                    del self.storage.vid_to_cached_value[vid]
                    del self.storage.vid_to_fingerprint[vid]
                    del self.storage.vid_to_result_metadata[vid]

    def remove_vid(self, vid: ValueIdentity):
        value = self.storage.vid_to_cached_value.get(vid)
        if value is not None:
            # TODO: add test cases for unlinking!!!
            with self.transaction_manager:
                del self.storage.vid_to_cached_value[vid]
                del self.storage.vid_to_fingerprint[vid]
                del self.storage.vid_to_result_metadata[vid]
                value.unlink()

    def get_vids(self):
        return set(self.storage.vid_to_cached_value.keys())

    def get_cached_value(self, vid: ValueIdentity):
        return self.storage.vid_to_cached_value.get(vid)

    def load_value(self, vid: ValueIdentity):
        cached_value = self.get_cached_value(vid)
        if not cached_value:
            return None

        # Load value
        with StopwatchContext() as stopwatch:
            loaded_value = cached_value.load()
            wrapped_value = MODULE_EXTENSIONS.wrap_return_value(loaded_value)

        metadata = self.get_result_metadata(vid)
        with self.transaction_manager:
            metadata.total_load_durations += stopwatch.elapsed_time
            metadata.num_loads += 1

        return wrapped_value

    def get_fingerprint(self, vid: ValueIdentity) -> Fingerprint:
        return self.storage.vid_to_fingerprint.get(vid)

    def get_result_metadata(self, vid: ValueIdentity) -> ResultMetadata:
        return self.storage.vid_to_result_metadata.get(vid)

    def tag(self, tag_name: str, vid: Optional[ValueIdentity]):
        with self.transaction_manager:
            self.storage.tag_to_vid.update(tag_name, vid)

    def get_tag_vid(self, tag_name) -> Optional[ValueIdentity]:
        return self.storage.tag_to_vid.get_value(tag_name)

    def get_tag_name(self, vid: ValueIdentity) -> Optional[str]:
        return self.storage.tag_to_vid.get_key(vid)

    def has_vid(self, vid):
        return vid in self.storage.vid_to_cached_value



from typing import TypeVar, MutableMapping, Dict, Tuple, Iterator

KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.
KT_co = TypeVar('KT_co', covariant=True)  # Value type covariant containers.
VT_co = TypeVar('VT_co', covariant=True)  # Value type covariant containers.


class KeyIdDict(MutableMapping[KT, VT]):
    id_key_value: Dict[int, Tuple[KT, VT]]

    def __init__(self):
        self.id_key_value = {}

    def __setitem__(self, k: KT, v: VT) -> None:
        self.id_key_value[id(k)] = (k, v)

    def __delitem__(self, k: KT) -> None:
        del self.id_key_value[id(k)]

    def __getitem__(self, k: KT) -> VT:
        return self.id_key_value[id(k)][1]

    def __len__(self) -> int:
        return len(self.id_key_value)

    def __iter__(self) -> Iterator[KT]:
        for key, value in self.id_key_value.values():
            yield key

    def __repr__(self):
        return f"KeyIdDict{{{', '.join(map(lambda key_value: f'{repr(key_value[0])}:{repr(key_value[1])}', self.id_key_value.values()))}}}"



from dataclasses import dataclass
from typing import Tuple, FrozenSet, Optional


MAX_FINGERPRINT_VALUE_LENGTH = 1024


class Fingerprint:
    pass


@dataclass(frozen=True)
class FingerprintDigest(Fingerprint):
    digest: object


@dataclass(frozen=True)
class FingerprintName(Fingerprint):
    name: str


@dataclass(frozen=True)
class FingerprintDigestRepr(FingerprintDigest):
    """`FingerprintDigest` that carries its original value to be more informative.

    For all purposes, we ignore the actual value for hashing and use the provided digest."""

    value: str

    def __eq__(self, other):
        return super().__eq__(other)

    def __hash__(self):
        return super().__hash__()


# We keep this separate from FunctionIdentity, so as to cache by identity
# and determine staleness using fingerprints.
# (Otherwise, we lack a key to index with and find stale entries.)
@dataclass(frozen=True)
class FunctionFingerprint(Fingerprint):
    fingerprint: object


# Includes dependencies.
@dataclass(frozen=True)
class DeepFunctionFingerprint(FunctionFingerprint):
    func_calls: FrozenSet[Tuple[Tuple[str, ...], FunctionFingerprint]]


class ResultFingerprint(Fingerprint):
    pass


@dataclass(frozen=True)
class CallFingerprint(ResultFingerprint):
    function: FunctionFingerprint
    # Need fingerprints everywhere! This needs to be a separate hierarchy!
    args: Tuple[Optional[Fingerprint], ...]
    kwargs: FrozenSet[Tuple[str, Optional[Fingerprint]]]


@dataclass(frozen=True)
class CellFingerprint:
    cell_code_fingerprint: FunctionFingerprint
    globals_load: FrozenSet[Tuple[Tuple[str, ...], Tuple["ValueIdentity", Fingerprint]]]
    outputs: FrozenSet[str]


@dataclass(frozen=True)
class CellResultFingerprint(ResultFingerprint):
    cell: CellFingerprint
    key: str





from typing import MutableSet, TypeVar, Iterator, Dict

T = TypeVar('T')
T_co = TypeVar('T_co', covariant=True)  # Any type covariant containers.


# TODO: add tests
class IdSet(MutableSet[T]):
    id_value: Dict[int, T]

    def __init__(self):
        self.id_value = {}

    def add(self, x: T) -> None:
        self.id_value[id(x)] = x

    def discard(self, x: T) -> None:
        if x in self:
            del self.id_value[id(x)]

    def clear(self) -> None:
        self.id_value.clear()

    def __contains__(self, x: object) -> bool:
        return id(x) in self.id_value

    def __len__(self) -> int:
        return len(self.id_value)

    def __iter__(self) -> Iterator[T_co]:
        return iter(self.id_value.values())

    def __repr__(self):
        return f"IdSet{{{ ', '.join(map(repr, self))}}}"





import weakref
from dataclasses import dataclass
from typing import MutableMapping, Set

from mamo.internal.bimap import MappingBimap
from mamo.internal.fingerprints import Fingerprint
from mamo.internal.common.key_id_dict import KeyIdDict
from mamo.internal.identities import ValueIdentity, ComputedValueIdentity
from mamo.internal.providers import ValueProvider
from mamo.internal.staleness_registry import StalenessRegistry
from mamo.internal.weakref_utils import WeakKeyIdMap


@dataclass
class AbstractValueRegistry(ValueProvider):
    vid_value_bimap: MappingBimap[ValueIdentity, object]
    value_fingerprint_map: MutableMapping[object, Fingerprint]
    staleness_registry: StalenessRegistry

    def identify_value(self, value) -> ValueIdentity:
        return self.vid_value_bimap.get_key(value)

    def fingerprint_value(self, value) -> Fingerprint:
        return self.value_fingerprint_map.get(value)

    def resolve_value(self, vid: ValueIdentity):
        return self.vid_value_bimap.get_value(vid)

    def resolve_fingerprint(self, vid: ValueIdentity):
        return self.fingerprint_value(self.resolve_value(vid))

    def add(self, vid: ValueIdentity, value, fingerprint: Fingerprint):
        assert value is not None

        if vid in self.vid_value_bimap:
            existing_value = self.vid_value_bimap.get_value(vid)
            self.staleness_registry.mark_stale(existing_value)
            del self.value_fingerprint_map[existing_value]

        self.vid_value_bimap.update(vid, value)
        self.value_fingerprint_map[value] = fingerprint
        self.staleness_registry.mark_used(value)

    def remove_vid(self, vid: ValueIdentity):
        value = self.vid_value_bimap.get_value(vid)
        if value is not None:
            self.remove_value(value)

    def remove_value(self, value: object):
        if not self.has_value(value):
            return

        self.staleness_registry.mark_stale(value)
        self.vid_value_bimap.del_value(value)
        del self.value_fingerprint_map[value]

    def has_vid(self, vid: ValueIdentity):
        return vid in self.vid_value_bimap

    def has_value(self, value):
        return self.vid_value_bimap.has_value(value)

    def get_vids(self) -> Set[ValueIdentity]:
        return set(self.vid_value_bimap.get_keys())


class ValueRegistry(AbstractValueRegistry):
    def __init__(self, staleness_registry: StalenessRegistry):
        vid_value_bimap = MappingBimap({}, KeyIdDict())
        value_to_fingerprint = KeyIdDict()
        super().__init__(vid_value_bimap, value_to_fingerprint, staleness_registry)


class WeakValueRegistry(AbstractValueRegistry):
    def __init__(self, staleness_registry: StalenessRegistry):
        vid_value_bimap = MappingBimap(weakref.WeakValueDictionary(), WeakKeyIdMap())
        value_to_fingerprint = WeakKeyIdMap()
        super().__init__(vid_value_bimap, value_to_fingerprint, staleness_registry)



from abc import ABC
from dataclasses import dataclass
from typing import Tuple, FrozenSet

from mamo.internal.fingerprints import Fingerprint, FingerprintName


class ValueIdentity:
    def get_external_info(self):
        raise NotImplementedError()


@dataclass(frozen=True)
class ValueFingerprintIdentity(ValueIdentity):
    qualified_type_name: str
    fingerprint: Fingerprint

    def get_external_info(self):
        if isinstance(self.fingerprint, FingerprintName):
            return f"{self.fingerprint.name}"
        return f"{self.qualified_type_name}"


def value_name_identity(unique_name: str):
    return ValueFingerprintIdentity("{value}", FingerprintName(unique_name))


@dataclass(frozen=True)
class FunctionIdentity:
    qualified_name: str


@dataclass(frozen=True)
class CellIdentity(FunctionIdentity):
    pass


class ComputedValueIdentity(ValueIdentity, ABC):
    pass


# TODO: merge this into CallIdentity?
@dataclass(frozen=True)
class ValueCallIdentity(ComputedValueIdentity):
    fid: FunctionIdentity
    args_vid: Tuple[ValueIdentity, ...]
    kwargs_vid: FrozenSet[Tuple[str, ValueIdentity]]

    def get_external_info(self):
        # TODO: maybe convert get_external_info into a visitor pattern?
        # TODO: add tests that test all of this?

        # Look at arguments.
        # Do we have any named ones?
        args = []
        for arg in self.args_vid:
            args.append(arg.get_external_info())
        for name, value in self.kwargs_vid:
            args.append(f"{name}={value.get_external_info()}")
        if args:
            args = ",".join(args)
        else:
            args = ""
        return f"{self.fid.qualified_name}({args})"


@dataclass(frozen=True)
class ValueCellResultIdentity(ComputedValueIdentity):
    cell: CellIdentity
    key: str

    def get_external_info(self):
        return f"cell_{self.cell.name}.{self.key}"


class ValueIdentityVisitor:
    def visit_fingerprint(self, vid: ValueFingerprintIdentity):
        raise TypeError(f"{type(vid)} not supported! (for {vid})")

    def visit_call(self, vid: ValueCallIdentity):
        raise TypeError(f"{type(vid)} not supported! (for {vid})")

    def visit_cell_result(self, vid: ValueCellResultIdentity):
        raise TypeError(f"{type(vid)} not supported! (for {vid})")

    def visit_computed_value(self, vid: ComputedValueIdentity):
        raise TypeError(f"{type(vid)} not supported! (for {vid})")

    def visit(self, vid: ValueIdentity):
        if isinstance(vid, ValueFingerprintIdentity):
            return self.visit_fingerprint(vid)
        elif isinstance(vid, ValueCallIdentity):
            return self.visit_call(vid)
        elif isinstance(vid, ValueCellResultIdentity):
            return self.visit_cell_result(vid)
        elif isinstance(vid, ComputedValueIdentity):
            return self.visit_computed_value(vid)
        raise NotImplementedError(f"Unknown type {type(vid)} for {vid}")



from IPython import get_ipython
from IPython.core.magic import Magics, magics_class, cell_magic
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring
import mamo


@magics_class
class MamoMagics(Magics):
    @cell_magic
    @magic_arguments()
    @argument("name", type=str, default=None, help="Name of the cell.")
    def mamo(self, line, cell_code):
        """mamo cell wrapper, only tracks global stores!"""
        assert isinstance(line, str)
        assert isinstance(cell_code, str)

        args = parse_argstring(self.mamo, line)
        mamo.run_cell(args.name, cell_code, self.shell.user_ns)


get_ipython().register_magics(MamoMagics)



import os
from abc import ABC
from dataclasses import dataclass, replace
from typing import Optional, NoReturn


class CachedValue:
    """Wraps a value that is being cached offline."""

    def unlink(self) -> NoReturn:
        # This value is about to not be part of the cache anymore.
        # Deal with it (by removing auxiliary files etc).
        pass

    def load(self):
        raise NotImplementedError()

    def get_stored_size(self) -> int:
        raise NotImplementedError()


@dataclass
class ExternallyCachedFilePath:
    """Builder for file paths for externally cached values."""

    path: str
    external_id: str
    vid_info: str

    @staticmethod
    def for_tuple_item(path: "Optional[ExternallyCachedFilePath]", i: int) -> "Optional[ExternallyCachedFilePath]":
        if path is None:
            return None

        return replace(path, external_id=path.external_id + f"_{i}")

    def build(self, cache_info, ext):
        return os.path.join(self.path, f"{self.vid_info}_{cache_info}_{self.external_id}.{ext}")


@dataclass
class ExternallyCachedValue(CachedValue, ABC):
    """A value that is cached with external resources."""

    path: str

    def unlink(self):
        unlinked_path = self.path + ".unlinked"
        # TODO: shall we pass the vid as argument and store it in a file next to
        # the unlinked entry?
        os.rename(self.path, unlinked_path)

    def get_stored_size(self):
        return os.path.getsize(self.path)





from mamo.internal import main

from mamo.internal.main import init_mamo

# TODO: what about exceptions?
# TODO: what about wrapping methods in class definitions?
# Shouldn't really support that maybe because we won't be able to perform
# code dependency checks for that...
mamo = main.Mamo.wrap_function


def _ensure_mamo_init():
    if main.mamo is None:
        print('Initializing Mamo!')
        init_mamo()


def _require_mamo():
    if main.mamo is None:
        raise RuntimeError("Mamo not initialized!")


def register_external_value(unique_name, value):
    _ensure_mamo_init()

    main.mamo.register_external_value(unique_name, value)


def tag(tag_name, value):
    _ensure_mamo_init()

    main.mamo.tag(tag_name, value)


def get_tag_value(tag_name):
    _ensure_mamo_init()

    return main.mamo.get_tag_value(tag_name)


def get_tag_name(value):
    _require_mamo()

    return main.mamo.get_tag_name(value)


def get_external_value(unique_name):
    _ensure_mamo_init()

    return main.mamo.get_external_value(unique_name)


def flush_online_cache():
    _require_mamo()
    main.mamo.flush_cache()


def flush_value(value):
    _require_mamo()
    main.mamo.flush_value(value)


def get_cached_value_identities(persisted=False):
    _ensure_mamo_init()

    # TODO: add require_mamo_init??? (so we don't accidentally initialize mamo in a getter
    vids = main.mamo.get_value_identities(persisted=persisted)

    # TODO: convert to dicts?
    return vids


def get_metadata(value):
    _require_mamo()

    return main.mamo.get_metadata(value)


def is_stale(value, *, depth=-1):
    _require_mamo()

    return main.mamo.is_stale(value, depth=depth)


def forget(value):
    _require_mamo()

    main.mamo.forget(value)


def run_cell(name, cell_code, namespace):
    _ensure_mamo_init()

    main.mamo.run_cell(name, cell_code, namespace)



import pickle
from dataclasses import dataclass

from mamo.internal.weakref_utils import ObjectProxy

from typing import Optional, Tuple

from mamo.internal.cached_values import ExternallyCachedFilePath, CachedValue, ExternallyCachedValue
from mamo.internal.db_stored_value import DBPickledValue
from mamo.internal.module_extension import ModuleExtension, ObjectSaver

import hashlib

from mamo.internal.reflection import get_type_qualified_name

MAX_PICKLE_SIZE = 2 ** 30


@dataclass
class CachedTuple(CachedValue):
    values: Tuple[CachedValue, ...]

    def load(self):
        return tuple(item.load() for item in self.values)

    def unlink(self):
        for item in self.values:
            item.unlink()

    def get_stored_size(self) -> int:
        return sum(item.get_stored_size() for item in self.values)


class DefaultTupleObjectSaver(ObjectSaver):
    def __init__(self, value, object_savers: Tuple[ObjectSaver]):
        super().__init__(value)
        self.object_savers = object_savers

    def get_estimated_size(self) -> int:
        return sum(object_saver.get_estimated_size() for object_saver in self.object_savers)

    def compute_digest_(self):
        hash_method = hashlib.md5()
        for object_saver in self.object_savers:
            hash_method.update(object_saver.compute_digest())
        return hash_method.digest()

    def cache_value(self, external_path_builder: Optional[ExternallyCachedFilePath]) -> Optional[CachedValue]:
        cached_items = tuple(
            object_saver.cache_value(ExternallyCachedFilePath.for_tuple_item(external_path_builder, i))
            for i, object_saver in enumerate(self.object_savers))
        if any(item is None for item in cached_items):
            return None
        return CachedTuple(cached_items)


class DefaultExternallyCachedValue(ExternallyCachedValue):

    def load(self):
        with open(self.path, "br") as external_file:
            pickled_bytes = external_file.read()

        try:
            return pickle.loads(pickled_bytes)
        except pickle.PickleError as err:
            # TODO: log err
            print(err)
            return None

    @staticmethod
    def save(external_path, pickled_bytes):
        # TODO: add error handling!
        with open(external_path, "bw") as external_file:
            external_file.write(pickled_bytes)

        return DefaultExternallyCachedValue(external_path)


class DefaultObjectSaver(ObjectSaver):
    def __init__(self, value, pickled_bytes):
        super().__init__(value)
        self.pickled_bytes = pickled_bytes

    def get_estimated_size(self) -> Optional[int]:
        return len(self.pickled_bytes)

    def compute_digest_(self):
        return hashlib.md5(self.pickled_bytes).digest()

    def cache_value(self, external_path_builder: Optional[ExternallyCachedFilePath]) -> Optional[CachedValue]:
        if external_path_builder is not None:
            external_path = external_path_builder.build(get_type_qualified_name(self.value), "pickle")
            cached_value = DefaultExternallyCachedValue.save(external_path, self.pickled_bytes)
        else:
            cached_value = DBPickledValue(self.pickled_bytes)

        # TODO: catch transactions error for objects that cannot be pickled here?
        return cached_value


class DefaultModuleExtension(ModuleExtension):
    def supports(self, value):
        return True

    def get_object_saver(self, value) -> Optional[ObjectSaver]:
        if isinstance(value, tuple):
            return DefaultTupleObjectSaver(value, tuple(self.module_registry.get_object_saver(item) for item in value))

        try:
            pickled_bytes = pickle.dumps(value)
        except pickle.PicklingError as err:
            # TODO: log err
            print(err)
            return None

        if len(pickled_bytes) > MAX_PICKLE_SIZE:
            # TODO: log
            return None

        return DefaultObjectSaver(value, pickled_bytes)

    def wrap_return_value(self, value):
        # Treat tuples different for functions returning multiple values.
        # We usually want to wrap them separately.
        if isinstance(value, tuple):
            return tuple(self.module_registry.wrap_return_value(item) for item in value)

        if not isinstance(value, ObjectProxy):
            return ObjectProxy(value)

        # If the value is already wrapped, create a new proxy.
        # This is necessary so that if we pass the same object through nested mamo functions
        # the result doesn't share identities (which is important for staleness!)
        return ObjectProxy(value.__subject__)



from mamo.internal.fingerprints import Fingerprint, ResultFingerprint
from mamo.internal.common.id_set import IdSet
from mamo.internal.providers import ValueProvider
from mamo.internal.staleness_registry import StalenessRegistry
from mamo.internal.value_registries import WeakValueRegistry
from mamo.internal.identities import ValueIdentity, ComputedValueIdentity
from mamo.internal.persisted_store import PersistedStore


class ResultRegistry(ValueProvider):
    values: IdSet
    online_registry: WeakValueRegistry
    persisted_store: PersistedStore

    def __init__(self, staleness_registry: StalenessRegistry, persisted_cache: PersistedStore):
        self.values = IdSet()
        self.online_registry = WeakValueRegistry(staleness_registry)

        self.persisted_store = persisted_cache

    def identify_value(self, value) -> ValueIdentity:
        return self.online_registry.identify_value(value)

    def fingerprint_value(self, value):
        return self.online_registry.fingerprint_value(value)

    def get_vids(self):
        return self.online_registry.get_vids()

    def flush(self):
        self.values.clear()

    def flush_value(self, value):
        self.values.remove(value)

    def add(self, vid: ComputedValueIdentity, value, fingerprint: ResultFingerprint):
        assert isinstance(vid, ComputedValueIdentity)
        assert isinstance(fingerprint, ResultFingerprint)
        assert value is not None

        existing_value = self.online_registry.resolve_value(vid)
        if existing_value is value:
            return

        if existing_value is not None:
            self.values.discard(existing_value)

        self.values.add(value)
        self.online_registry.add(vid, value, fingerprint)
        self.persisted_store.add(vid, value, fingerprint)

    def remove_vid(self, vid: ValueIdentity):
        assert isinstance(vid, ComputedValueIdentity)

        self.persisted_store.remove_vid(vid)

        value = self.online_registry.resolve_value(vid)
        if value is None:
            return

        self.values.discard(value)
        self.online_registry.remove_value(value)

    def remove_value(self, value: object):
        if not self.has_value(value):
            return

        self.persisted_store.remove_vid(self.identify_value(value))
        self.values.discard(value)
        self.online_registry.remove_value(value)

    # TODO: rename to something that makes clear it might be very expensive!!
    def resolve_value(self, vid: ComputedValueIdentity):
        value = self.online_registry.resolve_value(vid)
        if value is None and self.persisted_store.has_vid(vid):
            value = self.persisted_store.load_value(vid)
            fingerprint = self.persisted_store.get_fingerprint(vid)
            self.values.add(value)
            self.online_registry.add(vid, value, fingerprint)

        return value

    def resolve_fingerprint(self, vid: ComputedValueIdentity):
        fingerprint = self.online_registry.fingerprint_value(self.online_registry.resolve_value(vid))

        if fingerprint is None:
            fingerprint = self.persisted_store.get_fingerprint(vid)

        return fingerprint

    def has_vid(self, vid):
        return self.online_registry.has_vid(vid) or self.persisted_store.has_vid(vid)

    def has_value(self, value):
        return self.online_registry.has_value(value)



import weakref
from typing import Dict, Iterator, TypeVar, Generic
from typing import MutableMapping, MutableSet
import objproxies

KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.
KT_co = TypeVar('KT_co', covariant=True)  # Value type covariant containers.
VT_co = TypeVar('VT_co', covariant=True)  # Value type covariant containers.

T = TypeVar('T')
T_co = TypeVar('T_co', covariant=True)  # Any type covariant containers.


def supports_weakrefs(value):
    return type(value).__weakrefoffset__ != 0


class ObjectProxy(objproxies.ObjectProxy):
    __slots__ = ("__weakref__",)


class IdMapFinalizer(Generic[KT]):
    id_to_finalizer: Dict[int, weakref.finalize]

    def __init__(self):
        self.id_to_finalizer = {}

    def _finalizer(self, id_value, custom_handler):
        del self.id_to_finalizer[id_value]
        if custom_handler is not None:
            custom_handler(id_value)

    def __contains__(self, item):
        return id(item) in self.id_to_finalizer

    def __len__(self):
        return len(self.id_to_finalizer)

    def __iter__(self):
        # TODO: add a test that shows that this is necessary to avoid deletions
        # Take a snapshot of the keys. This will ensure that the dictionary will be stable during iteration.
        return iter(list(map(self.lookup_id, self.id_to_finalizer)))

    def lookup_id(self, id_value):
        return self.id_to_finalizer[id_value].peek()[0]

    def register(self, value: KT, custom_handler):
        if not supports_weakrefs(value):
            # TODO: log?
            return

        id_value = id(value)
        if id_value in self.id_to_finalizer:
            raise ValueError(f"{value} has already been added to the finalizer!")

        self.id_to_finalizer[id_value] = weakref.finalize(value, self._finalizer, id_value, custom_handler)

    def release(self, value: KT):
        id_value = id(value)
        finalizer = self.id_to_finalizer.get(id_value)
        if finalizer is not None:
            finalizer.detach()
            del self.id_to_finalizer[id_value]

    def clear(self):
        for finalizer in self.id_to_finalizer.values():
            finalizer.detach()
        self.id_to_finalizer.clear()

    def __del__(self):
        self.clear()


# TODO: add tests
class WeakIdSet(MutableSet[T]):
    id_map_finalizer: IdMapFinalizer

    def __init__(self):
        self.id_map_finalizer = IdMapFinalizer()

    def add(self, x: T) -> None:
        assert supports_weakrefs(x)
        if x not in self.id_map_finalizer:
            self.id_map_finalizer.register(x, None)

    def discard(self, x: T) -> None:
        self.id_map_finalizer.release(x)

    def __contains__(self, x: object) -> bool:
        return x in self.id_map_finalizer

    def __len__(self) -> int:
        return len(self.id_map_finalizer)

    def __iter__(self) -> Iterator[T]:
        return iter(self.id_map_finalizer)

    def __repr__(self):
        return f"WeakIdSet{{{ ', '.join(map(repr, self))}}}"


class WeakKeyIdMap(MutableMapping[KT, VT]):
    id_map_to_value: Dict[int, VT]
    id_map_finalizer: IdMapFinalizer

    def __init__(self):
        self.id_map_to_value = {}
        self.id_map_finalizer = IdMapFinalizer()

    def _release(self, id_value):
        del self.id_map_to_value[id_value]

    def __setitem__(self, k: KT, v: VT) -> None:
        assert supports_weakrefs(k)

        id_k = id(k)
        if id_k not in self.id_map_to_value:
            self.id_map_finalizer.register(k, self._release)
        self.id_map_to_value[id_k] = v

    def __delitem__(self, k: KT) -> None:
        del self.id_map_to_value[id(k)]
        self.id_map_finalizer.release(k)

    def __getitem__(self, k: KT) -> VT:
        return self.id_map_to_value[id(k)]

    def __len__(self) -> int:
        return len(self.id_map_to_value)

    def __iter__(self) -> Iterator[KT]:
        # TODO: add a test that shows that this is necessary to avoid deletions
        # Take a snapshot of the keys. This will ensure that the dictionary will be stable during iteration.
        return iter(self.id_map_finalizer)

    def __repr__(self):
        return f"KeyIdDict{{{', '.join(map(lambda key: f'{repr(key)}:{repr(self[key])}', self))}}}"


class AbstractWrappedValueMutableMapping(Generic[KT, VT, T], MutableMapping[KT, VT]):
    data: Dict[KT, T]

    def __init__(self):
        self.data = {}

    def value_to_store(self, v: VT) -> T:
        raise NotImplementedError()

    def store_to_value(self, v: T) -> VT:
        raise NotImplementedError()

    def __setitem__(self, k: KT, v: VT) -> None:
        self.data[k] = self.value_to_store(v)

    def __delitem__(self, v: KT) -> None:
        del self.data[v]

    def __getitem__(self, k: KT) -> VT:
        return self.store_to_value(self.data[k])

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterator[KT]:
        return iter(self.data)



import dis
import hashlib
import inspect
import marshal
from dataclasses import dataclass
from types import FunctionType, CodeType
import builtins

# Bytecode-extracted features. Independent of runtime and can be cached.
from typing import FrozenSet, Tuple, Optional, List, Set


@dataclass(frozen=True)
class FunctionDependencies:
    global_loads: FrozenSet[Tuple[str, ...]]
    global_stores: FrozenSet[str]
    func_calls: FrozenSet[Tuple[str, ...]]


def get_module_name(value):
    return value.__class__.__module__


def _get_qualified_name(func):
    # TODO: handle Jupyter notebooks?
    # In notebooks, __module__ will be "__main__".
    return f"{func.__module__}.{func.__qualname__}"


def get_type_qualified_name(value):
    return _get_qualified_name(type(value))


def get_func_qualified_name(func):
    return _get_qualified_name(func)


@dataclass
class CodeAnalyzer:
    instructions: List[dis.Instruction]
    index: int = 0

    def get_argval(self):
        return self.instructions[self.index].argval

    def check_opnames(self, *opnames):
        return self.instructions[self.index].opname in opnames

    def nest(self):
        return CodeAnalyzer(self.instructions, self.index)

    def unroll_op_stack(self, remainder_stacksize):
        instruction = self.instructions[self.index]
        # + 1 as we ignore the return value that will be pushed onto the stack.
        stack_size = 1 - dis.stack_effect(instruction.opcode, instruction.arg)

        self.index -= 1

        while self.index >= 0 and stack_size > remainder_stacksize:
            instruction = self.instructions[self.index]
            stack_size -= dis.stack_effect(instruction.opcode, instruction.arg)
            self.index -= 1

        return self

    def unroll_load(self, reversed_qualified_name):
        while self.index >= 0:
            instruction = self.instructions[self.index]
            if instruction.opname == "LOAD_GLOBAL":
                reversed_qualified_name.append(instruction.argval)
                return tuple(reversed(reversed_qualified_name))
            elif instruction.opname == "LOAD_ATTR":
                reversed_qualified_name.append(instruction.argval)
            else:
                break
            self.index -= 1

        # TODO: log?
        return None

    def unroll_load_method(self):
        if self.check_opnames('LOAD_METHOD'):
            qualified_name = [self.get_argval()]
            self.index -= 1
            qualified_name = self.unroll_load(qualified_name)
            return qualified_name

    def try_parse_load(self, global_loads):
        instruction = self.instructions[self.index]
        if instruction.opname == "LOAD_GLOBAL":
            qualified_name = [instruction.argval]

            # Now try to resolve attribute accesses.
            while self.index <= len(self.instructions):
                self.index += 1
                instruction = self.instructions[self.index]
                if instruction.opname in ("LOAD_ATTR", "LOAD_METHOD"):
                    qualified_name.append(instruction.argval)
                else:
                    break

            global_loads.add(tuple(qualified_name))
            return True

        return False

    def try_parse_call(self, called_funcs: Set[Tuple[str, ...]]):
        if self.check_opnames("CALL_FUNCTION", "CALL_FUNCTION_KW", "CALL_FUNCTION_EX"):
            qualified_name = self.nest().unroll_op_stack(1).unroll_load([])
            if qualified_name is not None:
                called_funcs.add(qualified_name)

            self.index += 1
            return True
        elif self.check_opnames('CALL_METHOD'):
            qualified_name = self.nest().unroll_op_stack(2).unroll_load_method()
            if qualified_name is not None:
                called_funcs.add(qualified_name)

            self.index += 1
            return True
        return False

    def try_parse_store(self, global_stores: Set[str]):
        if self.check_opnames("STORE_GLOBAL"):
            global_stores.add(self.get_argval())
            self.index += 1
            return True
        return False

    def analyze(self):
        called_funcs = set()
        global_stores = set()
        global_loads = set()

        while self.index < len(self.instructions):
            if self.try_parse_load(global_loads):
                pass
            elif self.try_parse_store(global_stores):
                pass
            elif self.try_parse_call(called_funcs):
                pass
            else:
                self.index += 1

        return FunctionDependencies(frozenset(global_loads), frozenset(global_stores), frozenset(called_funcs))

    @staticmethod
    def from_func_or_code(func_or_code: FunctionType):
        instructions = list(dis.get_instructions(func_or_code))
        return CodeAnalyzer(instructions)


def get_code_object_fingerprint(code_object: CodeType):
    # TODO: add cache?
    hash_method = hashlib.md5(code_object.co_code)
    # This seems to only output stable objects for version==2!
    hash_method.update(marshal.dumps(code_object.co_consts, 2))
    return hash_method.digest()


def get_func_fingerprint(func: FunctionType):
    return get_code_object_fingerprint(func.__code__)


def resolve_qualified_name(qualified_name: Tuple[str, ...], namespace: dict):
    resolved = namespace.get(qualified_name[0])
    for attr in qualified_name[1:]:
        if resolved is None:
            break
        resolved = getattr(resolved, attr)
    return resolved


def resolve_qualified_names(qualified_names: FrozenSet[Tuple[str, ...]], namespace: dict):
    resolved_dict = {}
    for qualified_name in qualified_names:
        resolved_dict[qualified_name] = resolve_qualified_name(qualified_name, namespace)
    return resolved_dict


def get_func_deps(func: FunctionType) -> FunctionDependencies:
    # We can cache the globals dependencies and the code hash.
    # Then we need to create fingerprints for all referenced global values
    # And figure out whether we want to include dependencies for functions that are being
    # called.

    deps = CodeAnalyzer.from_func_or_code(func).analyze()

    loads = {load for load in deps.global_loads if load[0] not in deps.global_stores}
    calls = {call for call in deps.func_calls if call[0] not in deps.global_stores}

    loads.difference_update(calls)

    return FunctionDependencies(frozenset(loads), frozenset(deps.global_stores), frozenset(calls))


def is_func_local(func, local_prefix: Optional[str]):
    module = inspect.getmodule(func)
    if module is None:
        # TODO: log? this is weird
        return True

    # Functions in the main module are local by nature (so we don't need a local_prefix to establish that).
    if module.__name__ == "__main__":
        return True

    if local_prefix is None:
        return False

    # Python's builtin module does not have a __file__ field.
    if not hasattr(module, "__file__"):
        return False

    return module.__file__.startswith(local_prefix)


def is_func_builtin(func):
    if inspect.isbuiltin(func):
        return True
    if inspect.ismemberdescriptor(func):
        return True
    if inspect.ismethoddescriptor(func):
        return True
    module = inspect.getmodule(func)
    return module is builtins
